package controller

import (
	"electricity_bill_calc/model"
	"electricity_bill_calc/repository"
	"electricity_bill_calc/response"
	"electricity_bill_calc/security"
	"electricity_bill_calc/service"
	"electricity_bill_calc/utils"
	"net/http"

	"github.com/gin-gonic/gin"
	"github.com/google/uuid"
	"github.com/jinzhu/copier"
	"github.com/shopspring/decimal"
)

type _ParkController struct {
	Router *gin.Engine
}

var ParkController *_ParkController

func InitializeParkController(router *gin.Engine) {
	ParkController = &_ParkController{
		Router: router,
	}
	ParkController.Router.GET("/parks", security.EnterpriseAuthorize, listAllParksUnderSessionUser)
	ParkController.Router.GET("/parks/:uid", security.ManagementAuthorize, listAllParksUnderSpecificUser)
	ParkController.Router.POST("/park", security.EnterpriseAuthorize, createNewPark)
	ParkController.Router.PUT("/park/:pid", security.EnterpriseAuthorize, modifyPark)
	ParkController.Router.GET("/park/:pid", security.EnterpriseAuthorize, fetchParkDetail)
	ParkController.Router.PUT("/park/:pid/enabled", security.EnterpriseAuthorize, changeParkEnableState)
	ParkController.Router.DELETE("/park/:pid", security.EnterpriseAuthorize, deleteSpecificPark)
}

func listAllParksUnderSessionUser(c *gin.Context) {
	result := response.NewResult(c)
	session, exists := c.Get("session")
	if !exists {
		result.Error(http.StatusUnauthorized, "用户会话无效。")
		return
	}
	userSession, ok := session.(*model.Session)
	if !ok {
		result.Failure(http.StatusInternalServerError, "内部缓存错误，需要重新登录。")
		return
	}
	parks, err := repository.ParkRepo.ListAllParkBelongsTo(userSession.Uid)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	result.Json(http.StatusOK, "已获取到指定用户下的园区。", gin.H{"parks": parks})
}

func listAllParksUnderSpecificUser(c *gin.Context) {
	result := response.NewResult(c)
	requestUserId := c.Param("uid")
	parks, err := repository.ParkRepo.ListAllParkBelongsTo(requestUserId)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	result.Json(http.StatusOK, "已获取到指定用户下的园区。", gin.H{"parks": parks})
}

type _ParkInfoFormData struct {
	Name     string              `json:"name" form:"name"`
	Region   *string             `json:"region" form:"region"`
	Address  *string             `json:"address" form:"address"`
	Contact  *string             `json:"contact" form:"contact"`
	Phone    *string             `json:"phone" from:"phone"`
	Area     decimal.NullDecimal `json:"area" from:"area"`
	Capacity decimal.NullDecimal `json:"capacity" from:"capacity"`
	Tenement decimal.NullDecimal `json:"tenement" from:"tenement"`
	Category int                 `json:"category" form:"category"`
	Submeter int                 `json:"submeter" form:"submeter"`
}

func createNewPark(c *gin.Context) {
	result := response.NewResult(c)
	session, exists := c.Get("session")
	if !exists {
		result.Error(http.StatusUnauthorized, "用户会话无效。")
		return
	}
	userSession, ok := session.(*model.Session)
	if !ok {
		result.Failure(http.StatusInternalServerError, "内部缓存错误，需要重新登录。")
		return
	}
	formData := new(_ParkInfoFormData)
	c.BindJSON(formData)
	newPark := new(model.Park)
	copier.Copy(newPark, formData)
	newPark.Id = uuid.New().String()
	newPark.UserId = userSession.Uid
	nameAbbr := utils.PinyinAbbr(newPark.Name)
	newPark.Abbr = &nameAbbr
	newPark.Enabled = true
	err := service.ParkService.SaveNewPark(*newPark)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	result.Created("新园区完成创建。")
}

func modifyPark(c *gin.Context) {
	result := response.NewResult(c)
	session, exists := c.Get("session")
	if !exists {
		result.Error(http.StatusUnauthorized, "用户会话无效。")
		return
	}
	userSession, ok := session.(*model.Session)
	if !ok {
		result.Failure(http.StatusInternalServerError, "内部缓存错误，需要重新登录。")
		return
	}
	requestParkId := c.Param("pid")
	formData := new(_ParkInfoFormData)
	c.BindJSON(formData)
	park, err := repository.ParkRepo.FetchParkDetail(requestParkId)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	if userSession.Uid != park.UserId {
		result.NotAccept("不能修改不属于自己的园区。")
		return
	}
	copier.Copy(park, formData)
	nameAbbr := utils.PinyinAbbr(formData.Name)
	park.Abbr = &nameAbbr
	err = service.ParkService.UpdateParkInfo(park)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	result.Updated("指定园区资料已更新。")
}

func fetchParkDetail(c *gin.Context) {
	result := response.NewResult(c)
	session, exists := c.Get("session")
	if !exists {
		result.Error(http.StatusUnauthorized, "用户会话无效。")
		return
	}
	userSession, ok := session.(*model.Session)
	if !ok {
		result.Failure(http.StatusInternalServerError, "内部缓存错误，需要重新登录。")
		return
	}
	requestParkId := c.Param("pid")
	park, err := repository.ParkRepo.FetchParkDetail(requestParkId)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	if userSession.Uid != park.UserId {
		result.NotAccept("不能访问不属于自己的园区。")
		return
	}
	result.Json(http.StatusOK, "已经获取到指定园区的信息。", gin.H{"park": park})
}

type _ParkStateFormData struct {
	Enabled bool `json:"enabled" form:"enabled"`
}

func changeParkEnableState(c *gin.Context) {
	result := response.NewResult(c)
	session, exists := c.Get("session")
	if !exists {
		result.Error(http.StatusUnauthorized, "用户会话无效。")
		return
	}
	userSession, ok := session.(*model.Session)
	if !ok {
		result.Failure(http.StatusInternalServerError, "内部缓存错误，需要重新登录。")
		return
	}
	requestParkId := c.Param("pid")
	formData := new(_ParkStateFormData)
	c.BindJSON(formData)
	err := service.ParkService.ChangeParkState(userSession.Uid, requestParkId, formData.Enabled)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
		return
	}
	result.Updated("指定园区的可用性状态已成功更新。")
}

func deleteSpecificPark(c *gin.Context) {
	result := response.NewResult(c)
	session, exists := c.Get("session")
	if !exists {
		result.Error(http.StatusUnauthorized, "用户会话无效。")
		return
	}
	userSession, ok := session.(*model.Session)
	if !ok {
		result.Failure(http.StatusInternalServerError, "内部缓存错误，需要重新登录。")
		return
	}
	requestParkId := c.Param("pid")
	err := service.ParkService.DeletePark(userSession.Uid, requestParkId)
	if err != nil {
		result.Error(http.StatusInternalServerError, err.Error())
	}
	result.Deleted("指定园区已成功删除。")
}
